"""
PrimalScheme: a primer3 wrapper for designing multiplex primer schemes
Copyright (C) 2020 Joshua Quick and Andrew Smith
www.github.com/aresti/primalscheme

This module contains MultiplexScheme and supporting classes.

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>
"""

import logging

from primalscheme.wrapper import design_primers, InsufficientPrimersError
from primalscheme.components import (
    CandidatePrimer,
    CandidatePrimerPair,
    reversor,
)

logger = logging.getLogger("primalscheme")


class MultiplexScheme:
    """A complete multiplex primer scheme."""

    def __init__(
        self,
        references,
        p3_global,
        target_overlap,
        step_distance,
        min_unique,
        prefix,
        progress_func=None,
    ):

        self.references = references
        self.p3_global = p3_global
        self.target_overlap = target_overlap
        self.step_distance = step_distance
        self.min_unique = min_unique
        self.prefix = prefix
        self.progress_func = progress_func

        # derived
        self.primary_ref = references[0]
        self.ref_len = len(self.primary_ref)
        self.amplicon_size_min = p3_global["PRIMER_PRODUCT_SIZE_RANGE"][0][0]
        self.amplicon_size_max = p3_global["PRIMER_PRODUCT_SIZE_RANGE"][0][1]
        self.amplicon_max_variation = self.amplicon_size_max - self.amplicon_size_min
        self.primer_max_size = p3_global["PRIMER_MAX_SIZE"]

        self.regions = []

    def design_scheme(self):
        """Design a multiplex primer scheme"""

        regions = []
        region_num = 0
        is_last = False

        while not is_last:
            region_num += 1
            prev = regions[-1].top_pair if regions else None
            prev_in_pool = regions[-2].top_pair if region_num > 2 else None

            # determine left limit
            if region_num == 1:
                left_limit = 0
            elif region_num == 2 or prev.left.start > prev_in_pool.right.start:
                # first region in second pool, or we have a gap. Force progress.
                left_limit = prev.left.end + 1
            else:
                # default case (just don't crash into prev in pool)
                left_limit = prev_in_pool.right.start + 1

            # determine slice start
            if region_num == 1:
                slice_start = 0
            else:
                insert_start = prev.right.end - self.target_overlap - 1
                slice_start = (
                    insert_start - self.primer_max_size - self.amplicon_max_variation
                )

                # if target overlap is impossible, take left_limit
                slice_start = max(slice_start, left_limit)

                # handle last region
                remaining_distance = self.ref_len - prev.right.start
                is_last = remaining_distance <= self.amplicon_size_max
                if is_last:
                    right_aligned = self.ref_len - self.amplicon_size_max
                    # if forced to choose, take a gap at the end
                    slice_start = min(slice_start, right_aligned)

            # create region, find primers
            region = Region(region_num, self, left_limit, slice_start)
            try:
                region.find_primers()
            except NoSuitablePrimersError as e:
                if region_num == 1:
                    raise e  # we've got nothing
                break
            regions.append(region)
            if self.progress_func:
                self.progress_func(region.top_pair.right.start, self.ref_len)

        self.regions = regions
        if self.progress_func:
            self.progress_func(self.ref_len, self.ref_len)


class Window:
    """
    A sliding window representing slice coordinates
    against the primary reference.
    """

    def __init__(self, scheme, left_limit, slice_start, right_limit=None):
        self.scheme = scheme
        self.left_limit = left_limit
        self.slice_start = slice_start
        self.right_limit = right_limit or len(scheme.primary_ref.seq)

        self._initial_slice_start = self.slice_start

        logger.debug(
            f"Window: left_limit {left_limit}, slice_start {slice_start}, "
            f"right_limit {self.right_limit}"
        )

        # check bounds
        if slice_start < left_limit or self.slice_end > self.right_limit:
            raise SliceOutOfBoundsError("The window slice is out of bounds.")

    def step_left(self):
        distance = self.scheme.step_distance
        if (self.slice_start - distance) < self.left_limit:
            logger.debug("Left limit reached")
            raise SliceOutOfBoundsError("Left window limit reached.")
        self.slice_start -= distance
        logger.debug(f"Stepping left to {self.slice_start}")

    def step_right(self):
        distance = self.scheme.step_distance
        if (self.slice_end + distance) > self.right_limit:
            logger.debug("Right limit reached")
            raise SliceOutOfBoundsError("Right window limit reached.")
        self.slice_start += distance
        logger.debug(f"Stepping right to {self.slice_start}")

    def reset_slice(self):
        self.slice_start = self._initial_slice_start

    @property
    def slice_end(self):
        return self.slice_start + self.scheme.amplicon_size_max

    @property
    def ref_slice(self):
        primary_ref = self.scheme.primary_ref.seq  # Seq object
        return str(primary_ref[self.slice_start : self.slice_end])


class Region(Window):
    """
    Region forming part of a tiling amplicon scheme.
    """

    def __init__(self, region_num, *args, **kwargs):
        self.region_num = region_num
        self.pool = 1 if self.region_num % 2 == 1 else 2
        self.candidate_pairs = []
        self.top_pair = None
        super().__init__(*args, **kwargs)  # init Window

        logger.debug(f"Region {region_num}, pool {self.pool}")

    def find_primers(self):
        """
        Try <-> step window logic:
        """
        exhausted_left = False
        while True:
            try:
                self._find_primers_for_slice()
                break
            except InsufficientPrimersError:
                if exhausted_left:
                    try:
                        # step right, retry
                        self.step_right()
                    except SliceOutOfBoundsError:
                        raise NoSuitablePrimersError("Right limit reached.")
                else:
                    try:
                        # step left, retry
                        self.step_left()
                    except SliceOutOfBoundsError:
                        self.reset_slice()
                        exhausted_left = True
            except NoSuitablePrimersError as e:
                # out of options
                raise e

    def _find_primers_for_slice(self):
        """Try to find sufficient primers for the current slice"""

        logger.debug(f"Finding primers for slice [{self.slice_start}:{self.slice_end}]")

        pairs = design_primers(
            self.ref_slice,
            self.scheme.p3_global,
            self.scheme.min_unique,
            offset=self.slice_start,
        )

        name = f"{self.scheme.prefix}_{self.region_num}"
        for i in range(len(pairs[0])):
            left = CandidatePrimer(
                pairs[0][i].seq,
                pairs[0][i].start,
                CandidatePrimer.Direction.left,
                f"{name}_LEFT",
                penalty=pairs[0][i].penalty,
            )
            right = CandidatePrimer(
                pairs[1][i].seq,
                pairs[1][i].start,
                CandidatePrimer.Direction.right,
                f"{name}_RIGHT",
                penalty=pairs[1][i].penalty,
            )
            left.align(self.scheme.references)
            right.align(self.scheme.references)
            self.candidate_pairs.append(CandidatePrimerPair(left, right))

        self._pick_pair()

    def _sort_candidate_pairs(self):
        """Sort the list of candidate pairs in place"""

        self.candidate_pairs.sort(
            key=lambda x: (reversor(x.mean_identity), x.combined_penalty)
        )

    def _pick_pair(self):
        self._sort_candidate_pairs()
        self.top_pair = self.candidate_pairs[0]

        logger.debug(f"Picked: {self.top_pair}")
        for alignment in self.top_pair.left.alignments:
            logger.debug(alignment[1])
        for alignment in self.top_pair.right.alignments:
            logger.debug(alignment[1])
        for i, pair in enumerate(self.candidate_pairs):
            logger.debug(
                f"Candidate pair {i}: "
                f"{pair.mean_identity} identity, "
                f"penalty {pair.combined_penalty}"
            )


class NoSuitablePrimersError(Exception):
    """No suitable primers found."""

    pass


class SliceOutOfBoundsError(Exception):
    """The requested start position would put the slice out of bounds."""

    pass



"""
Tests for command line interface (CLI)
"""
import os
import pytest
import re

import primalscheme.cli

from argparse import Namespace


def test_runas_module():
    """
    Can this package be run as a Python module?
    """
    exit_status = os.system("python -m primalscheme --help")
    assert exit_status == 0


def test_entrypoint():
    """
    Is entrypoint script installed? (setup.py)
    """
    exit_status = os.system("primalscheme --help")
    assert exit_status == 0


def test_mutliplex_command():
    """
    Is multiplex command available?
    """
    exit_status = os.system("primalscheme multiplex --help")
    assert exit_status == 0


def test_cli_fails_without_command():
    """
    Does CLI stop execution w/o a command argument?
    """
    with pytest.raises(SystemExit):
        primalscheme.cli.main()
        pytest.fail("CLI doesn't abort asking for a command argument")


def test_cli_fails_without_fasta(default_config):
    """
    Does CLI stop execution w/o a positional fasta arg?
    """
    with pytest.raises(SystemExit):
        args = ["multiplex"]
        primalscheme.cli.parse_arguments(args, default_config)


@pytest.mark.parametrize(
    "option, value",
    [
        ("--prefix", "myScheme"),
        ("--amplicon-size-min", "380"),
        ("--amplicon-size-max", "420"),
        ("--target-overlap", "0"),
        ("--min-unique", "3"),
        ("--max-candidates", "10"),
        ("--output-path", "./scheme"),
        ("--step-distance", "11"),
        ("--debug", None),
        ("--force", None),
    ],
)
def test_availability_of_cli_options(option, value, default_config):
    """
    Are options available?
    """
    args = ["multiplex", "some.fa", option]
    if value:
        args.append(value)
    parsed = primalscheme.cli.parse_arguments(args, default_config)
    assert isinstance(parsed, Namespace)


def test_cli_fails_with_invalid_option(chikv_input, default_config):
    """
    Does CLI stop execution with an invalid option?
    """
    with pytest.raises(SystemExit):
        args = ["multiplex", f"{str(chikv_input)}", "--force", "--badOption"]
        primalscheme.cli.parse_arguments(args, default_config)


@pytest.mark.parametrize(
    "option", ["-V", "--version"],
)
def test_cli_version_output(option, capsys, default_config):
    """
    Does CLI output a sensible version number for -V and --version?
    """
    try:
        primalscheme.cli.parse_arguments([option], default_config)
    except SystemExit:
        pass
    out, err = capsys.readouterr()
    assert re.match(r"^primalscheme \d{1,2}.\d{1,2}.\d{1,2}[a-zA-Z]{0,3}\d*$", out)


def test_force_not_required_when_output_path_does_not_exist(tmp_path):
    path = tmp_path / "output"
    primalscheme.cli.get_output_path(output_path=path)


def test_force_required_when_output_path_does_exist(tmp_path):
    path = tmp_path / "output"
    path.mkdir()
    with pytest.raises(IOError):
        primalscheme.cli.get_output_path(output_path=path)


def test_force_allows_existing_output_path(tmp_path):
    path = tmp_path / "output"
    path.mkdir()
    primalscheme.cli.get_output_path(output_path=path, force=True)


def test_existing_output_path_not_dir(tmp_path):
    path = tmp_path / "output"
    path.write_text("text")
    with pytest.raises(IOError):
        primalscheme.cli.get_output_path(output_path=path, force=True)


def test_too_short_fasta_size_vs_amplicon_size(
    input_fasta_short_500, default_config, tmp_path, caplog
):
    args = ["multiplex", str(input_fasta_short_500)]
    parsed = primalscheme.cli.parse_arguments(args, default_config)
    outpath = tmp_path / "output"
    output_path = primalscheme.cli.get_output_path(outpath)

    with pytest.raises(SystemExit):
        primalscheme.cli.multiplex(parsed, output_path)
    messages = [log.message for log in caplog.records]
    assert any("too short" in msg for msg in messages)



"""
PrimalScheme: a primer3 wrapper for designing multiplex primer schemes
Copyright (C) 2020 Joshua Quick and Andrew Smith
www.github.com/aresti/primalscheme

This module contains the MutilplexReporter object.
This object extends a MultiplexScheme object to provide reporting methods.

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>
"""

import logging
import os
import pickle

from Bio import SeqIO
from Bio.Graphics import GenomeDiagram
from Bio.SeqFeature import FeatureLocation, SeqFeature
from reportlab.lib import colors

from primalscheme.multiplex import MultiplexScheme

logger = logging.getLogger("primalscheme")


class MultiplexReporter(MultiplexScheme):
    """Reporting methods to extend MultiplexScheme"""

    def write_all(self, path):
        self.write_bed(path)
        self.write_pickle(path)
        self.write_tsv(path)
        self.write_refs(path)
        self.write_schemadelica_plot(path)

    @property
    def inserts(self):
        return [(r.top_pair.left.end + 1, r.top_pair.right.end) for r in self.regions]

    @property
    def gap_count(self):
        gaps = 0
        last_covered = self.inserts[0][1]
        for insert in self.inserts:
            if insert[0] > last_covered:
                gaps += 1
            last_covered = insert[1]
        return gaps

    @property
    def percent_coverage(self):
        covered_coords = set([x for insert in self.inserts for x in range(*insert)])
        return round(len(covered_coords) / self.ref_len * 100, 2)

    def write_bed(self, path):
        """
        Write BED format to file.
        """
        filepath = path / f"{self.prefix}.scheme.bed"
        logger.info(f"Writing {filepath}")

        with open(filepath, "w") as bedhandle:
            for r in self.regions:
                print(
                    *map(
                        str,
                        [
                            self.primary_ref.id,
                            r.top_pair.left.start,
                            r.top_pair.left.end + 1,  # BED end is 1-based
                            r.top_pair.left.name,
                            r.pool,
                            "+",
                        ],
                    ),
                    sep="\t",
                    file=bedhandle,
                )
                print(
                    *map(
                        str,
                        [
                            self.primary_ref.id,
                            r.top_pair.right.end,
                            r.top_pair.right.start + 1,
                            r.top_pair.right.name,
                            r.pool,
                            "-",
                        ],
                    ),
                    sep="\t",
                    file=bedhandle,
                )

    def write_tsv(self, path):
        filepath = path / f"{self.prefix}.tsv"
        logger.info(f"Writing {filepath}")
        with open(filepath, "w") as tsvhandle:
            print(
                *["name", "pool", "seq", "size", "%gc", "tm (use 65)"],
                sep="\t",
                file=tsvhandle,
            )
            for r in self.regions:
                left = r.top_pair.left
                right = r.top_pair.right
                print(
                    *map(
                        str,
                        [
                            left.name,
                            r.pool,
                            left.seq,
                            left.size,
                            "%.2f" % left.gc,
                            "%.2f" % left.tm,
                        ],
                    ),
                    sep="\t",
                    file=tsvhandle,
                )
                print(
                    *map(
                        str,
                        [
                            right.name,
                            r.pool,
                            right.seq,
                            right.size,
                            "%.2f" % right.gc,
                            "%.2f" % right.tm,
                        ],
                    ),
                    sep="\t",
                    file=tsvhandle,
                )

    def write_pickle(self, path):
        filepath = path / f"{self.prefix}.pickle"
        logger.info(f"Writing {filepath}")
        with open(filepath, "wb") as pickleobj:
            pickle.dump(self, pickleobj)

    def write_refs(self, path):
        filepath = path / f"{self.prefix}.reference.fasta"
        logger.info(f"Writing {filepath}")
        with open(filepath, "w"):
            SeqIO.write(self.references, filepath, "fasta")

    def apply_to_window(self, sequence, window_size, function, step=None):
        """
        Modified from
        https://github.com/biopython/biopython/blob/master/Tests/test_GenomeDiagram.py
        Apply function to windows of the given sequence.
        Returns a list of (position, value) tuples for fragments of the passed
        sequence of length window_size (stepped by step), calculated by the
        passed function.  Returned positions are the midpoint of each window.
        Arguments:
        - sequence - Bio.Seq.Seq object.
        - window_size - an integer describing the length of sequence
          to consider.
        - step - an integer describing the step to take between windows
          (default = window_size//2).
        - function - Method or function that accepts a Bio.Seq.Seq object
          as its sole argument and returns a single value.
        apply_to_window(sequence, window_size, function) ->
            [(int, float),(int, float),...]
        """
        seqlen = len(sequence)  # Total length of sequence to be used
        if step is None:  # Use half window-width or 1 if larger
            step = max(window_size // 2, 1)
        else:  # Use specified step, or 1 if greater
            step = max(step, 1)

        results = []  # Holds (position, value) results

        # Perform the passed function on as many windows as possible, short of
        # overrunning the sequence
        pos = 0
        while pos < seqlen - window_size + 1:
            # Obtain sequence fragment
            start = pos
            middle = (pos + window_size + pos) // 2
            end = pos + window_size
            fragment = sequence[start:end]
            # Apply function to the sequence fragment
            value = function(fragment)
            results.append((middle, value))  # Add results to list
            # Advance to next fragment
            pos += step

        # Use the last available window on the sequence, even if it means
        # re-covering old ground
        if pos != seqlen - window_size:
            # Obtain sequence fragment
            pos = seqlen - window_size
            start = pos
            middle = (pos + window_size + pos) // 2
            end = pos + window_size
            fragment = sequence[start:end]
            # Apply function to sequence fragment
            value = function(fragment)
            results.append((middle, value))  # Add results to list

        return results  # Return the list of (position, value) results

    def calc_gc(self, sequence):
        """
        Return gc content as fraction
        """
        g = sequence.count("G") + sequence.count("g")
        c = sequence.count("C") + sequence.count("c")
        return (g + c) / len(sequence)

    def write_schemadelica_plot(self, path):
        gd_diagram = GenomeDiagram.Diagram("Primer Scheme", track_size=0.15)
        primer_feature_set = GenomeDiagram.FeatureSet()

        # make the gc track
        window = 50
        gc_set = GenomeDiagram.GraphSet("GC content")
        graphdata1 = self.apply_to_window(self.primary_ref.seq, window, self.calc_gc)
        gc_set.new_graph(
            graphdata1,
            "GC content",
            style="line",
            color=colors.violet,
            altcolor=colors.purple,
        )
        gc_track = GenomeDiagram.Track(
            "GC content", height=1.5, greytrack=0, scale_largetick_interval=1e3
        )
        gc_track.add_set(gc_set)

        # make the primer track
        for r in self.regions:
            region = str(r.region_num)
            strand = 1 if r.region_num % 2 else -1

            fwd_feature = SeqFeature(
                FeatureLocation(
                    r.top_pair.left.start, r.top_pair.left.end, strand=strand
                )
            )
            rev_feature = SeqFeature(
                FeatureLocation(
                    r.top_pair.right.end, r.top_pair.right.start, strand=strand
                )
            )
            region_feature = SeqFeature(
                FeatureLocation(
                    r.top_pair.left.start, r.top_pair.right.start, strand=strand
                )
            )

            primer_color = colors.red
            region_color = colors.palevioletred

            primer_feature_set.add_feature(
                region_feature,
                color=region_color,
                name=region,
                label=True,
                label_position="middle",
                label_angle=0 if strand == 1 else -180,
            )
            primer_feature_set.add_feature(fwd_feature, color=primer_color, name=region)
            primer_feature_set.add_feature(rev_feature, color=primer_color, name=region)

        primer_track = GenomeDiagram.Track(name="Annotated Features", height=1)
        primer_track.add_set(primer_feature_set)

        gd_diagram.add_track(primer_track, 2)
        gd_diagram.add_track(gc_track, 1)

        rows = max(2, int(round(len(self.primary_ref) / 10000.0)))
        gd_diagram.draw(
            format="linear",
            pagesize=(300 * rows, 200 * rows),
            fragments=rows,
            start=0,
            end=len(self.primary_ref),
        )

        pdf_filepath = os.path.join(path, "{}.pdf".format(self.prefix))
        svg_filepath = os.path.join(path, "{}.svg".format(self.prefix))
        logger.info(f"Writing {pdf_filepath}")
        logger.info(f"Writing {svg_filepath}")
        gd_diagram.write(pdf_filepath, "PDF", dpi=300)
        gd_diagram.write(svg_filepath, "SVG", dpi=300)



import logging
import primer3

from collections import namedtuple

logger = logging.getLogger("primalscheme")


class InsufficientPrimersError(Exception):
    """Unable to find sufficient unique primers at the current cursor position."""

    pass


def design_primers(seq, p3_global, min_unique, offset=0):
    """Find primer pairs for a sequence slice."""

    SimplePrimer = namedtuple("SimplePrimer", "seq start penalty")

    p3_seq = {
        "SEQUENCE_TEMPLATE": seq,
        "SEQUENCE_PRIMER_PAIR_OK_REGION_LIST": [-1, -1, -1, -1],
        "SEQUENCE_INCLUDED_REGION": [-1, -1],
    }

    # Call primer3
    p3_output = primer3.bindings.designPrimers(p3_seq, p3_global)

    # Parse result
    pairs = ([], [])
    text_dir = ("LEFT", "RIGHT")

    for d in range(2):
        num_returned = p3_output[f"PRIMER_{text_dir[d]}_NUM_RETURNED"]
        for i in range(num_returned):
            seq = str(p3_output[f"PRIMER_{text_dir[d]}_{i}_SEQUENCE"])
            penalty = float(p3_output[f"PRIMER_{text_dir[d]}_{i}_PENALTY"])
            start = offset + int(p3_output[f"PRIMER_{text_dir[d]}_{i}"][0])
            pairs[d].append(SimplePrimer(seq, start, penalty))

    # If we don't have min unique left and right, then raise
    unique_count = min(len(set(pairs[0])), len(set(pairs[1])))
    logger.debug(f"Primer3 returned {unique_count} unique pairs")

    if unique_count < min_unique:
        logger.debug(f"Does not satisfy {min_unique} min unique")
        raise InsufficientPrimersError(
            f"Failed to find {min_unique} unique left or right primers."
        )

    return pairs



"""
PrimalScheme: a primer3 wrapper for designing multiplex primer schemes
Copyright (C) 2020 Joshua Quick and Andrew Smith
www.github.com/aresti/primalscheme

This module contains the CLI for PrimalScheme.
It is executed when the user runs 'primalscheme' after installation,
or 'python -m primalscheme'.

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>
"""

import json
import sys
import argparse
import logging

from pathlib import Path

from Bio import SeqIO
from Bio.Seq import Seq
from Bio.SeqRecord import SeqRecord

from primalscheme import __version__ as version
from primalscheme.multiplex import NoSuitablePrimersError
from primalscheme.reporting import MultiplexReporter

logger = logging.getLogger("primalscheme")


def main():
    """
    Entry point for primalscheme.
    Process args, set up logging and call the command function.
    """

    config = get_config()
    try:
        args = parse_arguments(sys.argv[1:], config)
    except ValueError as e:
        print(f"Error: {e}")
        sys.exit(1)

    try:
        output_path = get_output_path(args.output_path, force=args.force)
    except IOError as e:
        print(f"Error: {e}")
        sys.exit(1)

    setup_logging(output_path, debug=args.debug, prefix=args.prefix)

    for arg in vars(args):
        logger.debug("{}: {}".format(arg, str(vars(args)[arg])))

    # Run
    args.func(args, output_path)


def multiplex(args, output_path):
    """
    Multipex scheme command.
    """

    # Process FASTA input
    try:
        min_ref_size = args.amplicon_size_max * 2
        references = process_fasta(args.fasta, min_ref_size=min_ref_size)
    except ValueError as e:
        logger.error(f"Error: {e}")
        sys.exit(2)

    # Log references
    logger.info(f"Designing primers using reference: {references[0].id}")
    for ref in references[1:]:
        logger.info(f"Checking alignments against reference: {ref.id}")

    # Create scheme
    try:
        scheme = MultiplexReporter(
            references,
            args.primer3,
            args.target_overlap,
            args.step_distance,
            args.min_unique,
            args.prefix,
            progress_func=stdout_progress,
        )
        scheme.design_scheme()
    except NoSuitablePrimersError:
        # Unable to find primers for at least one region
        logger.error("Error: Unable to find suitable primers")
        sys.exit(10)
    except Exception as e:
        # Unexpected error
        logger.error(f"Error: {e}")
        raise e

    # Write outputs
    logger.info(
        f"All done! Scheme created with {len(scheme.regions)} regions, "
        f"{scheme.gap_count } gap{'' if scheme.gap_count == 1 else 's'}, "
        f"{scheme.percent_coverage}% coverage"
    )
    scheme.write_all(output_path)
    sys.exit(0)


def process_fasta(file_path, min_ref_size=None):
    """
    Parse and validate the fasta file.
    """

    references = []
    records = SeqIO.parse(file_path, "fasta")  # may raise

    # Remove gaps
    for record in records:
        ref = SeqRecord(
            Seq(str(record.seq).replace("-", "").upper()),
            id=record.id,
            description=record.id,
        )
        references.append(ref)

    # Check for no references
    if not references:
        raise ValueError("The input FASTA file does not contain any valid references.")

    # Check for too short references
    if min_ref_size and any(len(ref) < min_ref_size for ref in references):
        raise ValueError(
            "One or more of your references is too short. Based on your target "
            f"amplicon size, the minimum reference size is {min_ref_size} nt."
        )

    # Check for too many references
    if len(references) > 100:
        raise ValueError("A maximum of 100 reference genomes is currently supported.")

    # Check for max difference in size between references
    primary_ref = references[0]
    primary_ref_len = len(primary_ref)

    if any(abs(len(r) - primary_ref_len) > 500 for r in references):
        raise ValueError(
            "One or more of your references is too different in size to "
            "the primary (first) reference. The maximum difference is "
            "500 nt."
        )

    # Check for a valid alphabet
    VALID_ALPHABET = "AGCTacgt"
    for r in references:
        if any(base not in VALID_ALPHABET for base in set(r.seq)):
            raise ValueError(
                "One or more of your fasta sequences contain invalid "
                f"nucleotide codes. The supported alphabet is '{VALID_ALPHABET}'. "
                "Ambiguity codes and gaps are not currently supported."
            )

    return references


def setup_logging(output_path, debug=False, prefix="primalscheme"):
    """
    Setup logging output and verbosity.
    """

    logger.setLevel(logging.DEBUG if debug else logging.INFO)

    # File handler
    log_filepath = output_path / f"{prefix}.log"
    fh = logging.FileHandler(log_filepath)
    fh.setLevel(logging.DEBUG)
    fh_formatter = logging.Formatter(
        "%(asctime)s - %(name)s - %(levelname)s - %(message)s"
    )
    fh.setFormatter(fh_formatter)
    logger.addHandler(fh)

    # Stream handler STDOUT
    sh = logging.StreamHandler(sys.stdout)
    sh.setLevel(logging.INFO)
    sh_formatter = logging.Formatter("%(message)s")
    sh.setFormatter(sh_formatter)
    logger.addHandler(sh)

    logger.info(f"Writing log to {log_filepath}")


def get_output_path(output_path, force=False):
    """
    Check for an existing output dir, require --force to overwrite.
    Create dir, return path object.
    """
    path = Path(output_path)

    if path.exists() and not force:
        raise IOError("Directory exists add --force to overwrite")
    elif path.exists() and not path.is_dir():
        raise IOError("The output path is not a directory.")

    path.mkdir(exist_ok=True)
    return path


def get_config():
    config_file = Path(__file__).parent / "config.json"
    return json.loads(config_file.read_text())


def parse_arguments(args, config):
    """
    Parse command line arguments.
    """
    # Setup parsers
    parser = argparse.ArgumentParser(
        prog="primalscheme",
        description=("A primer3 wrapper for designing multiplex primer schemes."),
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
    )
    subparsers = parser.add_subparsers(title="[subcommands]", dest="command")
    subparsers.required = True

    parser_scheme = subparsers.add_parser("multiplex", help="Multiplex PCR scheme")

    # Set func and defaults from config
    parser_scheme.set_defaults(func=multiplex, **config)

    # Add arguments to parser
    parser_scheme.add_argument("fasta", help="FASTA file")
    parser_scheme.add_argument(
        "--prefix",
        help="Prefix used for primer names and output files " "(default: %(default)s)",
    )
    parser_scheme.add_argument(
        "--amplicon-size-min",
        type=positive_int,
        default=config["primer3"].get("PRIMER_PRODUCT_SIZE_RANGE")[0][0],
        help="Minimum amplicon size (default: %(default)i)",
    )
    parser_scheme.add_argument(
        "--amplicon-size-max",
        type=positive_int,
        default=config["primer3"].get("PRIMER_PRODUCT_SIZE_RANGE")[0][1],
        help="Maximum amplicon size (default: %(default)i)",
    )
    parser_scheme.add_argument(
        "--min-unique",
        type=positive_int,
        help="Minimum unique candiate pairs (default: %(default)i)",
    )
    parser_scheme.add_argument(
        "--max-candidates",
        type=positive_int,
        default=config["primer3"].get("PRIMER_NUM_RETURN"),
        help="Maximum candidate pairs (default: %(default)i)",
    )
    parser_scheme.add_argument(
        "--output-path", help="Output directory (default: %(default)s)"
    )
    parser_scheme.add_argument(
        "--target-overlap",
        type=positive_int,
        help="Target overlap size (default: %(default)i)",
    )
    parser_scheme.add_argument(
        "--step-distance",
        type=positive_int,
        help="Distance to step between find attempts (default: %(default)i)",
    )
    parser_scheme.add_argument("--debug", action="store_true", help="Verbose logging")
    parser_scheme.add_argument(
        "--force",
        action="store_true",
        help="Force output to an existing directory and overwrite output " "files.",
    )
    parser.add_argument(
        "-V", "--version", action="version", version=f"%(prog)s {version}"
    )

    # Generate args
    parsed = parser.parse_args(args)

    # Post parsing checks
    if parsed.amplicon_size_max < parsed.amplicon_size_min:
        raise ValueError("--amplicon-size-min cannot exceed --amplicon-size-max")

    # override primer3 config
    parsed.__dict__["primer3"].update(
        {
            "PRIMER_NUM_RETURN": parsed.max_candidates,
            "PRIMER_PRODUCT_SIZE_RANGE": [
                [parsed.amplicon_size_min, parsed.amplicon_size_max]
            ],
        }
    )

    return parsed


def positive_int(string):
    value = int(string)
    if value < 0:
        raise argparse.ArgumentTypeError("positive integer required.")
    return value


def stdout_progress(count, total):
    """Update a simple stdout progress bar"""
    bar_len = 60
    filled_len = int(round(bar_len * count / float(total)))

    percents = round(100.0 * count / float(total), 1)
    bar = "=" * filled_len + "-" * (bar_len - filled_len)

    sys.stdout.write(f"\r[{bar}] {percents}%")
    if count == total:
        sys.stdout.write("\n")
    sys.stdout.flush()


if __name__ == "__main__":
    main()



import pytest

from Bio.Seq import Seq

from primalscheme.cli import get_config, process_fasta
from primalscheme.multiplex import MultiplexScheme


@pytest.fixture(scope="session")
def default_chikv_scheme(chikv_input):
    return get_scheme(chikv_input)


def get_scheme(fasta, config=None):
    if not config:
        config = get_config()
    references = process_fasta(fasta)
    scheme = MultiplexScheme(
        references,
        config.get("primer3"),
        config.get("target_overlap"),
        config.get("step_distance"),
        config.get("min_unique"),
        config.get("prefix"),
    )
    scheme.design_scheme()
    return scheme


def no_collisions(regions):
    pool_ids = (1, 2)
    for pool_id in pool_ids:
        pool = [r for r in regions if r.pool == pool_id]
        last_coord = -1
        for r in pool:
            if r.top_pair.left.start <= last_coord:
                return False
            last_coord = r.top_pair.right.start
    return True


def test_chikv_scheme_has_no_gaps(default_chikv_scheme):
    regions = default_chikv_scheme.regions
    inserts = [(r.top_pair.left.end + 1, r.top_pair.right.end) for r in regions]
    covered_coords = set([x for insert in inserts for x in range(*insert)])
    all_coords = set(
        range(regions[0].top_pair.left.end + 1, regions[-1].top_pair.right.end)
    )

    assert all_coords.issubset(covered_coords)


def test_no_collisions_in_any_test_scheme(all_stored_inputs):
    scheme = get_scheme(all_stored_inputs)
    assert no_collisions(scheme.regions)


@pytest.mark.parametrize("amplicon_size", range(200, 801, 200))
def test_scheme_varying_amplicon_sizes(default_config, amplicon_size, chikv_input):
    config = default_config
    variation = int(0.1 * amplicon_size / 2)
    amplicon_size_min = amplicon_size - variation
    amplicon_size_max = amplicon_size + variation
    config["primer3"].update(
        {"PRIMER_PRODUCT_SIZE_RANGE": [[amplicon_size_min, amplicon_size_max]]}
    )
    scheme = get_scheme(chikv_input, config)

    regions = scheme.regions
    for region in regions[:4]:
        product_size = region.top_pair.product_size
        assert product_size <= amplicon_size_max
        assert product_size >= amplicon_size_min


def test_large_target_overlap_does_not_result_in_collision(default_config, chikv_input):
    config = default_config.update({"target_overlap": 300})
    scheme = get_scheme(chikv_input, config)

    assert no_collisions(scheme.regions)


def test_candidates_are_correctly_sorted(default_chikv_scheme):
    region = default_chikv_scheme.regions[0]
    identity = region.top_pair.mean_identity
    penalty = region.top_pair.combined_penalty
    for candidate in region.candidate_pairs:
        if candidate.mean_identity > identity:
            pytest.fail("Candidates are not sorted first by mean identity")
        elif (
            candidate.mean_identity == identity and candidate.combined_penalty < penalty
        ):
            pytest.fail("Candidates are not sorted second by penalty")
        identity = candidate.mean_identity
        penalty = candidate.combined_penalty


def test_left_primer_seq_matches_ref_slice(default_chikv_scheme):
    primary_ref = str(default_chikv_scheme.primary_ref.seq)
    left = default_chikv_scheme.regions[0].top_pair.left
    ref_slice = primary_ref[left.start : left.end + 1]

    assert len(left.seq) == left.size
    assert ref_slice == left.seq


def test_right_primer_seq_matches_ref_slice(default_chikv_scheme):
    primary_ref = str(default_chikv_scheme.primary_ref.seq)
    right = default_chikv_scheme.regions[0].top_pair.right
    ref_slice = str(Seq(primary_ref[right.end : right.start + 1]).reverse_complement())

    assert len(right.seq) == right.size
    assert ref_slice == right.seq



import pytest
import warnings

from Bio.SeqRecord import SeqRecord

with warnings.catch_warnings():
    warnings.simplefilter("ignore", PendingDeprecationWarning)
    from primalscheme.cli import process_fasta


def test_process_fasta_empty_input(input_fasta_empty):
    with pytest.raises(ValueError, match="does not contain any valid references"):
        process_fasta(input_fasta_empty)


def test_process_fasta_invalid_alphabet(input_fasta_invalid_alphabet):
    with pytest.raises(ValueError, match="invalid nucleotide codes"):
        process_fasta(input_fasta_invalid_alphabet)


def test_process_fasta_too_many_records(input_fasta_101_random_valid):
    with pytest.raises(ValueError, match="A maximum of 100"):
        process_fasta(input_fasta_101_random_valid)


def test_process_fasta_size_difference_over_500(input_fasta_size_difference_over_500,):
    with pytest.raises(ValueError, match="too different in size"):
        process_fasta(input_fasta_size_difference_over_500)


def test_process_fasta_returns_list_of_seq_records(input_fasta_5_random_valid):
    references = process_fasta(input_fasta_5_random_valid)
    assert len(references) == 5
    assert all(isinstance(r, SeqRecord) for r in references)


def test_process_fasta_chikv_demo(chikv_input):
    references = process_fasta(chikv_input)
    assert len(references) == 2


def test_process_fasta_gaps_removed(input_fasta_valid_with_gaps):
    references = process_fasta(input_fasta_valid_with_gaps)
    assert "-" not in references[0]


def test_process_fasta_too_short_input(input_fasta_short_500):
    with pytest.raises(ValueError, match="too short"):
        process_fasta(input_fasta_short_500, min_ref_size=900)



"""
PrimalScheme: a primer3 wrapper for designing multiplex primer schemes
Copyright (C) 2020 Joshua Quick and Andrew Smith
www.github.com/aresti/primalscheme

This module contains supporting scheme components.

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>
"""

import logging
import parasail

from enum import Enum

from primer3 import calcTm

logger = logging.getLogger("primalscheme")


class Primer(object):
    """A primer."""

    class Direction(Enum):
        left = 1
        right = -1

    def __init__(self, seq, start, direction):
        self.direction = direction
        self.seq = seq
        self.start = start

    def __str__(self):
        return f"{self.direction}:{self.seq}:{self.start}"

    @property
    def size(self):
        return len(self.seq)

    @property
    def end(self):
        if self.direction == Primer.Direction.left:
            return self.start + self.size - 1
        elif self.direction == Primer.Direction.right:
            return self.start - self.size + 1


class CandidatePrimer(Primer):
    """A candidate primer."""

    def __init__(self, seq, start, direction, name="", penalty=0):
        super().__init__(seq, start, direction)
        self.name = name
        self.penalty = penalty
        self.identity = 0
        self.tm = calcTm(self.seq, mv_conc=50, dv_conc=1.5, dntp_conc=0.6)
        self.gc = 100.0 * (seq.count("G") + seq.count("C")) / len(seq)
        self.alignments = []

    def align(self, references):
        for ref in references[1:]:
            # align against non-primary references
            alignment = get_alignment(self, ref)
            if alignment:
                self.alignments.append(alignment)

        # Calculate average percent identity
        if self.alignments:
            scores = [a[0] for a in self.alignments]
            self.identity = sum(scores) / len(scores)
            return


class CandidatePrimerPair(object):
    """A pair of candidate primers."""

    def __init__(self, left, right):
        self.left = left
        self.right = right

    def __str__(self):
        return f"{self.left} -:- {self.right}"

    @property
    def mean_identity(self):
        return (self.left.identity + self.right.identity) / 2

    @property
    def product_size(self):
        return self.right.start - self.left.start + 1

    @property
    def combined_penalty(self):
        return self.left.penalty + self.right.penalty


def get_alignment(primer, reference):
    """An seqan alignment of a primer against a reference."""

    MATRIX = parasail.matrix_create("ACGT", 2, -1)
    OPEN = 2
    EXTEND = 1
    IDENTITY_THRESHOLD = 0.7

    if primer.direction == Primer.Direction.left:
        ref = reference.seq
    elif primer.direction == Primer.Direction.right:
        ref = reference.reverse_complement().seq

    # Semi-Global, do not penalize gaps at beginning and end of s2/database
    trace = parasail.sg_dx_trace_striped_sat(
        str(primer.seq), str(ref), OPEN, EXTEND, MATRIX
    )
    traceback = trace.get_traceback()

    query_end = trace.end_query + 1
    ref_end = trace.end_ref + 1

    # Get alignment strings
    aln_query = traceback.query[ref_end - query_end : ref_end]
    cigar = traceback.comp[ref_end - query_end : ref_end]
    aln_ref = traceback.ref[ref_end - query_end : ref_end]

    # Identity for glocal alignment
    identity = cigar.count("|") / len(cigar)

    # Alignment failed
    if identity < IDENTITY_THRESHOLD:
        return None

    # Format alignment
    refid = reference.id[:30]
    name = primer.name[:30]
    formatted_query = f"{name: <30} {1: >6} {aln_query} {query_end}"
    if primer.direction == Primer.Direction.left:
        formatted_ref = (
            f"{refid: <30} {ref_end - query_end + 1: >6} {aln_ref} {ref_end}"
        )
    elif primer.direction == Primer.Direction.right:
        rev_start = len(reference) - ref_end
        formatted_ref = (
            f"{refid: <30} {rev_start + query_end: >6} {aln_ref} {rev_start + 1}"
        )
    formatted_cigar = f"{'': <30} {'': >6} {cigar}"
    formatted_alignment = "\n".join(
        ["", formatted_query, formatted_cigar, formatted_ref]
    )

    del trace

    return (identity, formatted_alignment)


class reversor:
    """Decorator to reverse sort comparisons"""

    def __init__(self, obj):
        self.obj = obj

    def __eq__(self, other):
        return other.obj == self.obj

    def __lt__(self, other):
        return other.obj < self.obj

    def __gt__(self, other):
        return other.obj > self.obj
